; **********************************************
; *         Temperature Controlled Fan         *
; *                ATtiny 25                   *
; *       NTC at ADC2-ADC3 (diff mode)         *
; *   PWM out at PB1 (OC1A), shared with MISO  *
; *                                            *
; *  Timer/Counter0: used for schedule (ADC)   *
; *  Timer/Counter1: used for PWM output,      *
; *                  asynchronous mode (64MHz) *
; *                                            *
; *                                            *
; *           (C)2015 by Ingo Barke            *
; **********************************************
;
;

; Output frequency for PWM pin:
; if HI_SPEED_PWM is defined: 64Mhz PWM (-> 250kHz @ 8 Bit and 8MHz),
; otherwise: 8MHz PWM (-> 31.25kHz @ 8 Bit and 8MHz)
.equ HI_SPEED_PWM = 1

;
; ============================================
; H A R D W A R E    I N F O R M A T I O N
; ============================================
;
;
.equ F_CPU = 8000000 ; CPU clock in Hz

; ============================================
; P O R T S   A N D   P I N S
; ============================================
;
; ADC Inputs:
; PB0 is connected to internal reference (bypass capacitor)
.equ ADCPortIn = PINB
.equ ADCDDR = DDRB
.equ ADCPin0 = 3 ; 1st voltage pin
.equ ADCPin1 = 4 ; 2nd voltage pin

; PWM Port:
.equ PWMPort = PORTB
.equ PWMDDR = DDRB
.equ PWMPin = 1 ; PWM output pin, shared with MISO. NOTE: Cannot be changed arbitrarily!


;
; ============================================
; C H A N G E A B L E   V A L U E S
; ============================================
;

; Timer0 compare match:
.equ T0_OC = 7 ; determines interrupt frequency, e.g. for reading keys
	; (7 -> 8 periods -> 3906.25Hz).

; Number of Timer0 interrupts until ADC conversion is done:
.equ ADC_CYC = 20 ; must be chosen such that ADC time per conversion is >= 13 ADC cycles.
	; (20 -> 195.3125Hz)

; Oversampling bits, i.e. 2^ADC_OS conversions done for each PWM update.
; Maximum value = 6 (10bit ADC + 6 bits)
.equ ADC_OS_BITS = 4 ; 16 conversions (4 -> 12.2Hz)

; Output calculation:
.equ ADC_LO = 200 ; ADC threshold below which output is at maximum
.equ ADC_HI = 500 ; ADC threshold above which output is at minimum
.equ PWM_MIN = 0 ; minimum PWM output value (8bit)
.equ PWM_MAX = 255 ; maximum PWM output value (8bit)




;
; ============================================
; C O N S T A N T S   A N D 
; D E R I V E D   V A L U E S
; ============================================
;

; Slope for output calculation (only applies if ADC value is in between ADC_LO and ADC_HI):
; out = PWM_MAX - ((((adcSum >> ADC_OS) - ADC_LO) * PWM_SLOPE) >> 8)
.equ PWM_SLOPE = ((PWM_MAX - PWM_MIN) * 256) / (ADC_HI - ADC_LO) ; 16bit prefactor: 256 = unity slope

; Oversampling factor, i.e. ADC_OS conversions done for each PWM update.
; Maximum value = 64 (10bit ADC + 6 bits)
.equ ADC_OS = (1<<ADC_OS_BITS) ; calculate from oversampling bits

; ============================================
; R E G I S T E R   D E F I N I T I O N S
; ============================================
;


.def buf1 = r2 ; short-term buffer
.def buf2 = r3 ; short-term buffer

.def adcSumL = r4 ; contains sum of ADC values (Lo byte)
.def adcSumH = r5 ; contains sum of ADC values (Hi byte)

.def adcValL = r6 ; contains current ADC value (Lo byte)
.def adcValH = r7 ; contains current ADC value (Hi byte)

.def zeroReg = r15 ; zero for 16bit arithmetics

.def temp1 = r16 ; multipurpose register
.def temp2 = r17 ; multipurpose register
.def temp3 = r18 ; multipurpose register
.def temp4 = r19 ; multipurpose register
.def temp5 = r20 ; multipurpose register

.def osCnt = r21 ; oversampling counter: if zero, PWM output is updated
.def adcCnt = r22 ; ADC counter: if zero, next ADC conversion is started

.def jobFlag = r23 ; job flags for interrupt routines

.def tempL = r24 ; 16 bit multipurpose register, Lo Byte
.def tempH = r25 ; 16 bit multipurpose register, Hi Byte

;
; ============================================
; S R A M   D E F I N I T I O N S
; ============================================
;
.DSEG
.ORG 0X0062
; Format:
; Lable: .BYTE N

;
; ============================================
; E E P R O M   D E F I N I T I O N S
; ============================================
;
;.ESEG
;.ORG $0000
; Format:
; Lable: .BYTE N


;
; ============================================
; R E S E T   A N D   I N T V E C T O R S
; ============================================
;
.CSEG
.ORG $0000
	rjmp Main ; Reset Vektor 1
	reti ; Int Vektor 2
	reti ; Int Vektor 3
	reti ; Int Vektor 4
	reti ; Int Vektor 5
	reti ; Int Vektor 6
	reti ; Int Vektor 7
	reti ; Int Vektor 8
	reti ; Int Vektor 9
	reti ; Int Vektor 9
	rjmp isr_timer0_CMA ; Int Vektor 11: Timer/Counter0 Compare Match A
	reti ; Int Vektor 12
	reti ; Int Vektor 13
	reti ; Int Vektor 14
	reti ; Int Vektor 15


; ============================================
; I N C L U D E S
; ============================================
;

; .include "avr200.asm" ; multiplication and division routines

;
; ============================================
;  M A C R O   D E F I N I T I O N S
; ============================================
;



; 16bit x 16bit = 32bit unsigned multiplication,
;
; Needs 7 registers (6 lower and 1 upper).
;
; From avr200.asm:
;	mc16uH, mc16uL = multiplicant (remains unchanged)
;	mp16uH, mp16uL = multiplier (overwritten by two lowest result bytes)
;	m16u3, m16u2, m16u1, m16u0 = result (two low bytes share registers with multiplier)
;	mcnt16u = loop counter (overwritten with zero)
;
; All registers are arbritary (0-31), except for mcnt16 (last one), which must be >=16 (ldi).
; Needs about 155 cycles, depending on number of zeros in multiplier.
;
; Labels in macros are always local (i.e relative labels generated by toolchain)

.MACRO MUL_16_16_32 ; m16u3, m16u2, m16u1/mp16uH, m16u0/mp16uL, mc16uH, mc16uL, mcnt16u
		clr	@0 ; clear 2 highest bytes of result
		clr	@1
		ldi	@6, 16 ; init loop counter
		lsr	@2
		ror	@3
	MUL_16_16_32_1:
		brcc MUL_16_16_32_2 ; if bit 0 of multiplier set
		add	@1, @5 ; add multiplicand Low to byte 2 of res
		adc	@0, @4 ; add multiplicand high to byte 3 of res
	MUL_16_16_32_2:
		ror	@0 ; shift right result byte 3
		ror	@1 ; rotate right result byte 2
		ror	@2 ; rotate result byte 1 and multiplier High
		ror	@3 ; rotate result byte 0 and multiplier Low
		dec	@6 ; decrement loop counter
		brne MUL_16_16_32_1 ; if not done, loop more
.ENDMACRO


.LISTMAC ; show macro in list file (for debug purposes)

; ============================================
; M A I N   P R O G R A M :   I N I T
; ============================================
;
Main:
	cli; interrups aus

	; Init stack:
	ldi temp1, LOW(RAMEND) ; Hi-byte of upper RAM address
	out SPL, temp1
	
	; Registers:
	clr zeroReg ; init zero register
	ldi osCnt, ADC_OS ; init oversampling counter
	ldi adcCnt, ADC_CYC ; init ADC counter

	; RAM:

	; PWM:
	in temp1, PWMPort
	andi temp1, ~(1<<PWMPin) ; off (Low)
	out PWMPort, temp1

	in temp1, PWMDDR
	ori temp1, (1<<PWMPin)
	out PWMDDR, temp1
	
	; ADC:
	in temp1, ADCDDR
	andi temp1, ~((1<<ADCPin0) | (1<<ADCPin0)) ; inputs to Lo, otherwise unchanged
	out ADCDDR, temp1

		; Register ADMUX:
		; REFS2-0 = 1 1 1 for int. reference 2.56V and bypass cap at PB0
		; ADLAR = 0 fuer rechtsbuendig (besser fuer 10bit)
		; MUX3-0 = 0110 (diff input PB3 (-) and PB4 (+); gain = 1)
	ldi temp1, (1<<REFS2) | (1<<REFS1) | (1<<REFS0) | (1<<MUX2) | (1<<MUX1)
	out ADMUX, temp1

		; Register ADCSRA:
		; ADEN: Enable (1)
		; ADSC: Start conversion (1)
		; ADFR: Free running mode (0)
		; ADIE: AD interrupt enable (0)
		; ADPS2-0 = 110 for prescaler = 64 (125kHz @ 8MHz, ADC_clock should be 50-200kHz)
	ldi temp1, (1<<ADEN) | (1<<ADSC) | (1<<ADPS2) | (1<<ADPS1)
	out ADCSRA, temp1

	rcall getVoltage ; dummy conversion

	; Timer0:
		; Register TCCR0A:
		; WGM02-0 = 010 (CTC mode; note that WGM2 bit is in TCCR0B register)
	ldi temp1, (1<<WGM01)
	out TCCR0A, temp1

		; Register TCCR0B:
		; CS02-0 = 100 for prescaler = 256 (31kHz @ 8MHz)
		; WGM02 = 0 (see TCCR0A register)
	ldi temp1, (1<<CS02)
	out TCCR0B, temp1

		; Register OCR0A:
		; set output compare match value:
	ldi temp1, T0_OC
	out OCR0A, temp1

		; Register TIMSK (shared with timer1):
		; OCIE0A: Output compare match interrupt enable (1)
	ldi temp1, (1<<OCIE0A)
	out TIMSK, temp1

	; Timer1 (PWM):
	ldi temp1, 255 ; set OCR1C to 255; necessary for 8bit PWM
	out OCR1C, temp1

	ldi temp1, 0 ; set output to zero
	out OCR1A, temp1

		; Register TCCR1:
		; PWM1A: enable PWM mode (1)
		; COM1A1-0: 01 (OC1x set at 0 and cleared on compare match -> positive PWM)
		; CS13-0: 0001 for prescaler = 1 (PCK or CK; max. frequency)
	ldi temp1, (1<<PWM1A) | (1<<COM1A0) | (1<<CS10)
	out TCCR1, temp1

		; Register TIMSK (shared with timer0)
		; no timer1 interrupt -> TIMSK remains unchanged

		.ifdef HI_SPEED_PWM ; start PLL if Hi-speed PWM is selected
			; Register PLLCSR (fast asynchronous mode, 64MHz):
			; action required:
			;	1) enable PLL (via PLLE bit)
			;	2) wait 100us; then wait until PLOCK bit is set (then PLL is stable)
			;	3) enable PCK (via PCKE bit) for selecting PCK as clock source for timer1
			ldi temp1, (1<<PLLE) ; step 1)
			out PLLCSR, temp1
			; wait at least 100us:
			ldi tempH, (100 * F_CPU / 1000000 / 256 / 4) + 1 ; at least 100us
			clr tempL
			PLL_wait0:
				sbiw tempH:tempL, 1 ; 2 cycles
				brne PLL_wait0 ; 2 cycles if not zero
			PLL_wait1:
				in temp1, PLLCSR ; load PLLCSR register
				sbrs temp1, PLOCK ; check if PLOCK bit is set
				rjmp PLL_wait1 ; not yet set
			; now enable PLL as clock source:
			in temp1, PLLCSR
			ori temp1, (1<<PCKE) ; set clock source
			out PLLCSR, temp1
		.endif


	; Sleep mode und interrupts:
	clr jobFlag ; init jobFlag
	sei ; enable interrupts

		; Register MCUCR:
		; SE: sleep mode enabled (1)
	ldi temp1, (1<<SE)
	out MCUCR, temp1


; ============================================
; M A I N   P R O G R A M   L O O P
; ============================================
;
mainLoop:
	tst jobFlag ; interrupt occured?
	brne loop_intActive
	sleep ; sleep
	nop ; dummy for wake-up
	nop
	loop_intActive:
		sbrs jobFlag, 0 ; bit 0 set?
		rjmp loop_noJob0 ; if not, go to end
			cbr jobFlag, 1<<0 ; delete jobFlag
			dec adcCnt ; decrease ADC counter
			brne loop_noJob0 ; no further action if not yet zero
				ldi adcCnt, ADC_CYC ; reset counter
				rcall nextADC ; read ADC value and start next conversion
		loop_noJob0:
rjmp mainLoop ; back to sleep

; ============================================
; S U B P R O G R A M S
; ============================================
;

; Reads voltage and stores ADC value in adcValL and adcValH
getVoltage:
	sbi ADCSRA, ADSC ; start ADC
	getVoltage_ADC_wait:
		sbic ADCSRA, ADSC ; bit will be cleared when ADC finished
	rjmp getVoltage_ADC_wait

	; read ADC result:
	in adcValL, ADCL ; always read Lo byte first...
	in adcValH, ADCH ; ...then read Hi byte (meanwhile locked)
	
	ret

; Reads current ADC value, updates adcSum, and starts next ADC conversion.
; If oversampling is finished (osCnt is zero), PWM output value is updated.
nextADC:
	; read ADC result:
	in adcValL, ADCL ; always read Lo byte first...
	in adcValH, ADCH ; ...then read Hi byte (meanwhile locked)

	add adcSumL, adcValL ; add to oversampling sum
	adc adcSumH, adcValH

	dec osCnt ; decrease counter
	breq nextADC_updatePWM ; if zero, reset counter and update PWM
		rjmp nextADC_noPWMUpdate ; continue without PWM update

	nextADC_updatePWM:
		ldi osCnt, ADC_OS ; init counter
		
		; divide by 2^ADC_OS_BITS:
		; first add half of ADC_OS for rounding:
		ldi temp1, ADC_OS / 2
		add adcSumL, temp1
		adc adcSumH, zeroReg
		; now start division (cycles: 4 or 7 + 5 * (ADC_OS_BITS - 1) = 4, 7, 12, 17,..., max. 32)
		ldi temp1, ADC_OS_BITS ; init counter ([+1])
		tst temp1 ; ([+1])
		breq nextADC_divide_1 ; ([+2]/[+1])
		nextADC_divide_0:
			lsr adcSumH ; divide by 2 (1 cyc [+1])
			ror adcSumL ; (1 cyc [+1])
			dec temp1 ; decrease counter (1 cyc [+1])
			brne nextADC_divide_0 ; (2 cyc [+1]]
		nextADC_divide_1:

		; now calculate PWM output:
		; first check if adcSum is within limits:
		ldi temp1, low(ADC_LO) ; load value for lower ADC limit
		ldi temp2, high(ADC_LO)
		sub adcSumL, temp1 ; subtraction: adcSum = adcSum - ADC_LO
		sbc adcSumH, temp2
		brsh nextADC_valAboveLo ; if ADC value is above ADC_LO, continue with upper bound
			ldi temp1, PWM_MAX ; load max. PWM value...
			rjmp nextADC_updatePWMValue ; ... and proceed with PWM update
		
		nextADC_valAboveLo:
		ldi temp1, low(ADC_HI - ADC_LO) ; load value for upper minus lower ADC limit
		ldi temp2, high(ADC_HI - ADC_LO)
		cp adcSumL, temp1 ; check if adcSum is beyond upper bound
		cpc adcSumH, temp2
		brlo nextADC_valWithinRange ; if ADC value is within PWM range, continue with output calculation
			ldi temp1, PWM_MIN ; load min. PWM value...
			rjmp nextADC_updatePWMValue ; ... and proceed with PWM update

		nextADC_valWithinRange:
		; calculate output: out = PWM_MAX - ((((adcSum >> ADC_OS) - ADC_LO) * PWM_SLOPE) >> 8),
		;	where ((adcSum >> ADC_OS) - ADC_LO) is already available in adcSumH:L
		ldi tempL, low(PWM_SLOPE) ; load slope value (multiplicant)
		ldi tempH, high(PWM_SLOPE)
		mov temp2, adcSumL ; load multiplicator (will be overwritten)
		mov temp3, adcSumH
		; call multiplication macro: result = PWM_SLOPE * adcSum
		MUL_16_16_32 temp5, temp4, temp3, temp2, tempH, tempL, temp1
		ldi temp1, PWM_MAX
		sub temp1, temp3 ; temp3 contains multiplication result divided by 256
		
		nextADC_updatePWMValue:
		out OCR1A, temp1 ; update PWM value for compare match
		clr adcSumL ; reset ADC sum value
		clr adcSumH

	nextADC_noPWMUpdate:
	sbi ADCSRA, ADSC ; start ADC

	ret


;
; ============================================
; I N T E R R U P T   S E R V I C E S
; ============================================
;
;

isr_timer0_CMA: ; timer0 compare match A
	push temp1 ; save register
	in temp1, SREG ; save status register
	push temp1
	
	; set jobFlag:
	sbr jobFlag, (1<<0) ; changes status register

	pop temp1 ; get status register
	out SREG, temp1 ; restore status register
	pop temp1 ; restore register
	reti



; ============================================
; D A T A   S T O R A G E
; ============================================
; Strings, text, tables, etc.
; Format: 
; Label:
; .db "   example text   ",0


;
; End of source code
;
.db "(C)2015 by IB " ; human
.db "C(5)10 2ybI  B" ; word wise



