; **********************************************************
; *               Temperature Controlled Fan               *
; *                      ATtiny 25                         *
; *                                                        *
; *    NTC: ADC2 (=PB4), single ended (relative to GND)    *
; * Regulator input: ADC3 (=PB3), sgl. ended (rel.to GND)  *
; *         PWM out: PB1 (OC1A), shared with MISO          *
; *                                                        *
; *    Timer/Counter0: used for global scheduling (ADC)    *
; *        Timer/Counter1: used for PWM output,            *
; *                        asynchronous mode (64MHz)       *
; *                                                        *
; *                                                        *
; *                   (C)2015 Ingo Barke                   *
; **********************************************************

; NTC is measured; from this value the output setpoint value (in ADC units)
; is calculated. The actual output voltage is measured at a second ADC input;
; the difference between this value and the setpoint is used as error for a
; feedback loop (integral regulator). The PWM output is varied such that actual
; output voltage is equal to the setpoint voltage. PWM limits can be defined.
; Low NTC voltages (i.e. high temperatures) are converted to high setpoint
; voltages (i.e. high fan speed).
;
; Timing:
; Timer0 compare match defines interrupt frequency,
; adcCnt is decremented ADC_CYC times, then ADC conversion is done.
; When adcCnt is zero, the next ADC channel is selected (adcCh)
; Each channel is measured 2^ADC_OS_BITS times (oversampling), the result
; is divided by 2^ADC_OS_BITS (rounded average value). Then the feedback
; parameters are updated.

; Output frequency for PWM pin:
; if HI_SPEED_PWM is defined: 64Mhz PWM (-> 250kHz @ 8 Bit and 8MHz),
; if not (i.e. if commented out): 8MHz PWM (-> 31.25kHz @ 8 Bit and 8MHz)
.equ HI_SPEED_PWM = 1


; ============================================
; H A R D W A R E    I N F O R M A T I O N
; ============================================

.equ F_CPU = 8000000 ; CPU clock in Hz

; ============================================
; P O R T S   A N D   P I N S
; ============================================
;

; Number of ADC channels (i.e. measured voltages):
.equ ADCNumChannels = 2

; ADC Inputs:
; PB0 is connected to internal reference (bypass capacitor)
.equ ADCPortIn = PINB
.equ ADCDDR = DDRB
.equ ADCPinNTC = 4 ; NTC input pin (ADC2)
.equ ADCPinReg = 3 ; regulator input pin (ADC3)

; PWM Port:
.equ PWMPort = PORTB
.equ PWMDDR = DDRB
.equ PWMPin = 1 ; PWM output pin, shared with MISO. NOTE: Cannot be changed arbitrarily!



; ============================================
; C H A N G E A B L E   V A L U E S
; ============================================


; Timer0 compare match:
.equ T0_OC = 3 ; determines interrupt frequency, e.g. for reading keys
	; (3 -> 4 periods -> 7812.5Hz @ 8MHz and prescaler=256, i.e. 31.25kHz clock).
	; => about 1000 CPU cycles avaiable for each timer0 interrupt.

; Number of Timer0 interrupts until ADC conversion is done:
.equ ADC_CYC = 2 ; must be chosen such that ADC time per conversion is >= 13 ADC cycles.
	; (2 -> 3906.25Hz)

; Oversampling bits, i.e. 2^ADC_OS conversions done for each PWM update.
; For good stability PWM update period should be longer than circuit time constant.
; Maximum value = 6 (10bit ADC + 6 bits)
.equ ADC_OS_BITS = 5 ; 32 conversions (5 -> 122Hz for all channels, i.e. 61Hz for each)

; Output calculation:
.equ NTC_LO = 120 ; ADC threshold for NTC below which setpoint is at maximum
.equ NTC_HI = 240 ; ADC threshold for NTC above which setpoint is at minimum
.equ SP_MIN = 110 ; minimum setpoint voltage for output (in ADC units)
.equ SP_MAX = 600 ; maximum setpoint voltage for output (in ADC units)

.equ PWM_MIN = 0 ; minimum PWM output value (8bit)
.equ PWM_MAX = 255 ; maximum PWM output value (8bit)

; Regulator parameters:
; I: integral weight: output = ( SUM(REG_I * error) ) >> 16 )
; max. 16bit (2^16-1)
.equ REG_I = 8000



; ============================================
; C O N S T A N T S   A N D 
; D E R I V E D   V A L U E S
; ============================================


; Slope for setpoint calculation (only applies if ADC value is in between ADC_LO and ADC_HI):
; setpoint = SP_MAX - ((((adcSum >> ADC_OS) - ADC_LO) * SP_SLOPE) >> 8)
.equ SP_SLOPE = ((SP_MAX - SP_MIN) * 256) / (NTC_HI - NTC_LO) ; 16bit prefactor: 256 = unity slope

; Oversampling factor, i.e. ADC_OS conversions done for each setpoint and PWM update.
; Maximum value = 64 (10bit ADC + 6 bits)
.equ ADC_OS = (1<<ADC_OS_BITS) ; calculate from oversampling bits

.equ DELAY_100US = (100 * F_CPU / 1000000) ; number of cycles that corresponds to 100us

; ============================================
; R E G I S T E R   D E F I N I T I O N S
; ============================================

.def adcValL = r1 ; current ADC value (any channel) (Lo byte)
.def adcValH = r2 ; current ADC value (any channel) (Hi byte)

.def adcSum0L = r3 ; current ADC sum for channel 0 (NTC) (Lo byte)
.def adcSum0H = r4 ; current ADC sum for channel 0 (NTC) (Hi byte)
.def adcSum1L = r5 ; current ADC sum for channel 1 (reg. in) (Lo byte)
.def adcSum1H = r6 ; current ADC sum for channel 1 (reg. in) (Hi byte)

.def regI0 = r7 ; current integral value of regulator, byte 0 (lowest)
.def regI1 = r8 ; current integral value of regulator, byte 1
.def regI2 = r9 ; current integral value of regulator, byte 2
.def regI3 = r10 ; current integral value of regulator, byte 3 (highest)

.def buf0 = r11 ; temporary buffer
.def buf1 = r12 ; temporary buffer
.def bufL = r13 ; temporary buffer for 16bit arithmetics (Lo byte)
.def bufH = r14 ; temporary buffer for 16bit arithmetics (Hi byte)

.def zeroReg = r15 ; zero for 16bit arithmetics

.def temp1 = r16 ; multipurpose register
.def temp2 = r17 ; multipurpose register
.def temp3 = r18 ; multipurpose register
;.def temp4 = r19 ; multipurpose register
;	.def debReg = r19 ; debug register

.def adcCnt = r20 ; ADC counter: if zero, next ADC conversion is started
.def adcCh = r21 ; counter for ADC channels: if zero, oversampling is continued
.def osCnt = r22 ; oversampling counter: if zero, setpoint/PWM voltage is updated

.def jobFlag = r23 ; job flags for interrupt routines

.def setpointL = r24 ; current setpoint value (Lo byte)
.def setpointH = r25 ; current setpoint value (Hi byte)


; ============================================
; S R A M   D E F I N I T I O N S
; ============================================

.DSEG
.ORG 0X0062
; Format:
; Label: .BYTE N


; ============================================
; E E P R O M   D E F I N I T I O N S
; ============================================
;
;.ESEG
;.ORG $0000
; Format:
; Lable: .BYTE N



; ============================================
; R E S E T   A N D   I N T V E C T O R S
; ============================================

.CSEG
.ORG $0000
	rjmp Main ; Reset Vektor 1
	reti ; Int Vektor 2
	reti ; Int Vektor 3
	reti ; Int Vektor 4
	reti ; Int Vektor 5
	reti ; Int Vektor 6
	reti ; Int Vektor 7
	reti ; Int Vektor 8
	reti ; Int Vektor 9
	reti ; Int Vektor 9
	rjmp isr_timer0_CMA ; Int Vektor 11: Timer/Counter0 Compare Match A
	reti ; Int Vektor 12
	reti ; Int Vektor 13
	reti ; Int Vektor 14
	reti ; Int Vektor 15


; ============================================
; I N C L U D E S
; ============================================


; ============================================
;  M A C R O   D E F I N I T I O N S
; ============================================

; Labels in macros are always local (i.e relative labels generated by toolchain)


; 16bit x 16bit = 32bit unsigned multiplication.
;
; Needs 7 registers (6 arbitrary and 1 upper).
;
; From avr200.asm:
;	mc16uH, mc16uL = multiplicand (remains unchanged)
;	mp16uH, mp16uL = multiplier (overwritten by two lowest result bytes)
;	m16u3, m16u2, m16u1, m16u0 = result (two low bytes share registers with multiplier)
;	mcnt16u = loop counter (overwritten with zero)
;
; All registers are arbritary (0-31), except for mcnt16u (last one), which must be >=16 (ldi).
; Needs about 148-164 cycles, depending on number of zeros in multiplier.
.MACRO MUL_16_16_32 ; m16u3, m16u2, m16u1/mp16uH, m16u0/mp16uL, mc16uH, mc16uL, mcnt16u
		clr	@0 ; clear 2 highest bytes of result
		clr	@1
		ldi	@6, 16 ; init loop counter
		lsr	@2
		ror	@3
	MUL_16_16_32_1:
		brcc MUL_16_16_32_2 ; if bit 0 of multiplier set
		add	@1, @5 ; add multiplicand Low to byte 2 of res
		adc	@0, @4 ; add multiplicand high to byte 3 of res
	MUL_16_16_32_2:
		ror	@0 ; shift right result byte 3
		ror	@1 ; rotate right result byte 2
		ror	@2 ; rotate result byte 1 and multiplier High
		ror	@3 ; rotate result byte 0 and multiplier Low
		dec	@6 ; decrement loop counter
		brne MUL_16_16_32_1 ; if not done, loop more
.ENDMACRO


; 16bit x 16bit = 32bit signed multiplication.
;
; Needs 7 registers (6 arbitrary and 1 upper).
;
; From avr200.asm:
;	mc16sH, mc16sL = multiplicand (remains unchanged)
;	mp16sH, mp16sL = multiplier (overwritten by two lowest result bytes)
;	m16s3, m16s2, m16s1, m16s0 = result (two low bytes share registers with multiplier)
;	mcnt16s = loop counter (overwritten with zero)
;
; All registers are arbritary (0-31), except for mcnt16s (last one), which must be >=16 (ldi).
; Needs 210-226 cycles, depending on number of zeros in multiplier.
.MACRO MULS_16_16_32 ; m16s3, m16s2, m16s1/mp16sH, m16s0/mp16sL, mc16sH, mc16sL, mcnt16s
	clr	@0 ; clear result byte 3
	sub	@1, @1 ; clear result byte 2 and carry
	ldi	@6, 16 ; init loop counter
	MULS_16_16_32_1:
		brcc MULS_16_16_32_2 ; if carry (previous bit) set...
		add	@1, @5 ; ...add multiplicand Low to result byte 2
		adc	@0, @4 ; add multiplicand High to result byte 3
	MULS_16_16_32_2:
		sbrc @3, 0 ;if current bit set...
		sub	@1, @5 ; ...sub multiplicand Low from result byte 2
		sbrc @3, 0 ;if current bit set...
		sbc	@0, @4 ; ...sub multiplicand High from result byte 3
		asr	@0 ; shift right result and multiplier
		ror	@1
		ror	@2
		ror	@3
		dec	@6; decrement counter
		brne MULS_16_16_32_1 ; if not done, loop more	
.ENDMACRO

; Negates signed 16-bit number (two's complement)
; NOT USED !!!
; Needs 2 high registers (16-31) due to subi/sbci.
; Needs 4 cycles.
.MACRO NEG_16 ; valH, valL
	com @1 ;Invert Lo byte
	com	@0 ;Invert Hi byte
	subi @1, -1	; Add 1
	sbci @0, 0
.ENDMACRO

.LISTMAC ; show macros in list file (for debug purposes)

; ============================================
; M A I N   P R O G R A M :   I N I T
; ============================================

Main:
	cli; interrups aus

	; Init stack:
	ldi temp1, LOW(RAMEND) ; Hi-byte of upper RAM address
	out SPL, temp1
	
	; Registers:
	clr jobflag ; clear flags
	clr zeroReg ; init zero register
	ldi osCnt, ADC_OS ; init oversampling counter
	ldi adcCnt, ADC_CYC ; init ADC counter
	ldi adcCh, ADCNumChannels ; init channel counter
	ldi setpointL, LOW(SP_MIN) ; init setpoint: minimum value
	ldi setpointH, HIGH(SP_MIN)
	clr regI0 ; clear integral value
	clr regI1 ; clear integral value
	clr regI2 ; clear integral value
	clr regI3 ; clear integral value

	; PWM:
	in temp1, PWMPort
	andi temp1, ~(1<<PWMPin) ; off (Low)
	out PWMPort, temp1

	in temp1, PWMDDR
	ori temp1, (1<<PWMPin)
	out PWMDDR, temp1
	
	; ADC:
	in temp1, ADCDDR
	andi temp1, ~((1<<ADCPinNTC) | (1<<ADCPinReg)) ; inputs to Lo, otherwise unchanged
	out ADCDDR, temp1

		; Register ADMUX:
		; REFS2-0 = 1 1 1 for int. reference 2.56V and bypass cap at PB0
		; ADLAR: Left adjust result (0)
		; MUX3-0 = 0010 (single ended input ADC2 at PB4)
	ldi temp1, (1<<REFS2) | (1<<REFS1) | (1<<REFS0) | (1<<MUX1)
	out ADMUX, temp1

		; Register ADCSRA:
		; ADEN: Enable (1)
		; ADSC: Start conversion (1)
		; ADFR: Free running mode (0)
		; ADIE: AD interrupt enable (0)
		; ADPS2-0 = 110 for prescaler = 64 (125kHz @ 8MHz, ADC_clock should be 50-200kHz)
	ldi temp1, (1<<ADEN) | (1<<ADSC) | (1<<ADPS2) | (1<<ADPS1)
	out ADCSRA, temp1

	rcall getVoltage ; dummy conversion

	; Timer0:
		; Register TCCR0A:
		; WGM02-0 = 010 (CTC mode; note that WGM2 bit is in TCCR0B register)
	ldi temp1, (1<<WGM01)
	out TCCR0A, temp1

		; Register TCCR0B:
		; CS02-0 = 100 for prescaler = 256 (31kHz @ 8MHz)
		; WGM02 = 0 (see TCCR0A register)
	ldi temp1, (1<<CS02)
	out TCCR0B, temp1

		; Register OCR0A:
		; set output compare match value:
	ldi temp1, T0_OC
	out OCR0A, temp1

		; Register TIMSK (shared with timer1):
		; OCIE0A: Output compare match interrupt enable (1)
	ldi temp1, (1<<OCIE0A)
	out TIMSK, temp1

	; Timer1 (PWM):
	ldi temp1, 255 ; set OCR1C to 255; necessary for 8bit PWM
	out OCR1C, temp1

	ldi temp1, 0 ; set output to zero
	out OCR1A, temp1

		; Register TCCR1:
		; PWM1A: enable PWM mode (1)
		; COM1A1-0: 01 (OC1x set at 0 and cleared on compare match -> positive PWM)
		; CS13-0: 0001 for prescaler = 1 (PCK or CK; max. frequency)
	ldi temp1, (1<<PWM1A) | (1<<COM1A0) | (1<<CS10)
	out TCCR1, temp1

		; Register TIMSK (shared with timer0)
		; no timer1 interrupt -> TIMSK remains unchanged

		.ifdef HI_SPEED_PWM ; start PLL if Hi-speed PWM is selected
			; Register PLLCSR (fast asynchronous mode, 64MHz):
			; action required:
			;	1) enable PLL (via PLLE bit)
			;	2) wait 100us; then wait until PLOCK bit is set (then PLL is stable)
			;	3) enable PCK (via PCKE bit) for selecting PCK as clock source for timer1
			ldi temp1, (1<<PLLE) ; step 1)
			out PLLCSR, temp1
			; wait at least 100us:
			ldi temp1, LOW(DELAY_100US / 4 + 1)
			ldi temp2, HIGH(DELAY_100US / 4 + 1)
			PLL_wait0:
				subi temp1, 1 ; 1 cycle
				sbci temp2, 0 ; 1 cycle
				brne PLL_wait0 ; 2 cycles if not zero
			PLL_wait1:
				in temp1, PLLCSR ; load PLLCSR register
				sbrs temp1, PLOCK ; check if PLOCK bit is set
				rjmp PLL_wait1 ; not yet set
			; now enable PLL as clock source:
			in temp1, PLLCSR
			ori temp1, (1<<PCKE) ; set clock source
			out PLLCSR, temp1
		.endif


	; Sleep mode und interrupts:
	clr jobFlag ; init jobFlag
	sei ; enable interrupts

		; Register MCUCR:
		; SE: sleep mode enabled (1)
	ldi temp1, (1<<SE)
	out MCUCR, temp1

	; Debug:
		; clr debReg

; ============================================
; M A I N   P R O G R A M   L O O P
; ============================================

mainLoop:
	tst jobFlag ; interrupt occured?
	brne loop_intActive
	sleep ; sleep
	nop ; dummy for wake-up
	nop
	loop_intActive:
		sbrs jobFlag, 0 ; bit 0 set?
		rjmp loop_noJob0 ; if not, go to end
			cbr jobFlag, 1<<0 ; delete jobFlag
			dec adcCnt ; decrease ADC counter
			brne loop_noJob0 ; no further action if not yet zero
				ldi adcCnt, ADC_CYC ; reset counter
				rcall nextADC ; read ADC value and start next conversion
		loop_noJob0:
rjmp mainLoop ; back to sleep

; ============================================
; S U B P R O G R A M S
; ============================================

; Reads voltage and stores ADC value in adcValL and adcValH
getVoltage:
	sbi ADCSRA, ADSC ; start ADC
	getVoltage_ADC_wait:
		sbic ADCSRA, ADSC ; bit will be cleared when ADC finished
	rjmp getVoltage_ADC_wait

	; read ADC result:
	in adcValL, ADCL ; always read Lo byte first...
	in adcValH, ADCH ; ...then read Hi byte (meanwhile locked)
	
	ret

; Reads current ADC value, updates adcSum, and starts next ADC conversion.
; If oversampling is finished, setpoint and PWM output are updated.
nextADC:
	dec adcCh ; decrease channel counter
	
	; read ADC result:
	in adcValL, ADCL ; always read Lo byte first...
	in adcValH, ADCH ; ...then read Hi byte (meanwhile locked)

	in temp1, ADMUX ; read ADMUX register
	andi temp1, 0b11110000 ; clear MUX3-0 bits, channel will be set in following code

	cpi adcCh, 0 ; channel 0 (NTC)?
	brne nextADC_calcSum_not0
		add adcSum0L, adcValL ; add to oversampling sum
		adc adcSum0H, adcValH
		; set ADC channel (now adcCh=0 -> next conversion on adcCh=1 -> ADC3):
		ori temp1, 0b00000011 ; ADC2 -> MUX3-0=0010 
		out ADMUX, temp1 ; set channel
		rjmp nextADC_calcSum_finished
	nextADC_calcSum_not0:
		cpi adcCh, 1 ; channel 1 (regulator input)?
		brne nextADC_calcSum_not1 ; should never happen (for easy extension)
			add adcSum1L, adcValL ; add to oversampling sum
			adc adcSum1H, adcValH
			; set ADC channel (now adcCh=1 -> next conversion on adcCh=0 -> ADC2):
			ori temp1, 0b00000010 ; ADC3 -> MUX3-0=0011 
			out ADMUX, temp1 ; set channel
			rjmp nextADC_calcSum_finished
		nextADC_calcSum_not1:
			rjmp nextADC_calcSum_not1 ; freeze (should never happen)

	nextADC_calcSum_finished:

	sbi ADCSRA, ADSC ; start ADC already now, ensures (almost) constant ADC timing

	cpi osCnt, 1 ; last oversampling cycle before data evaluation?
	breq nextADC_action ; if so, change setpoint/PWM output
		rjmp nextADC_noAction ; if not, continue conversion sequence
	
	nextADC_action:
		cpi adcCh, 0 ; channel 0 (NTC)?
		breq nextADC_action_0
			rjmp nextADC_action_not0
		nextADC_action_0:
			; Update setpoint
			; 
			; divide sumVal by 2^ADC_OS_BITS:
			; first add half of ADC_OS for rounding:
			ldi temp1, ADC_OS / 2
			add adcSum0L, temp1
			adc adcSum0H, zeroReg
			; now start division (cycles: 4 or 7 + 5 * (ADC_OS_BITS - 1) = 4, 7, 12, 17,..., max. 32)
			ldi temp1, ADC_OS_BITS ; init counter ([+1])
			tst temp1 ; ([+1])
			breq nextADC_Ch0_divide_1 ; ([+2]/[+1])
			nextADC_Ch0_divide_0:
				lsr adcSum0H ; divide by 2 (1 cyc [+1])
				ror adcSum0L ; (1 cyc [+1])
				dec temp1 ; decrease counter (1 cyc [+1])
				brne nextADC_Ch0_divide_0 ; (2 cyc [+1]]
			nextADC_Ch0_divide_1:

			; now calculate setpoint:
			; first check if adcSum is within limits:
			ldi temp1, LOW(NTC_LO) ; load value for lower ADC limit (for NTC)
			ldi temp2, HIGH(NTC_LO)
			sub adcSum0L, temp1 ; subtraction: adcSum = adcSum - ADC_LO
			sbc adcSum0H, temp2
			brsh nextADC_Ch0_valAboveLo ; if ADC value is above ADC_LO, continue with upper bound
				ldi setpointL, LOW(SP_MAX) ; load max. setpoint value
				ldi setpointH, HIGH(SP_MAX)
				rjmp nextADC_Ch0_clearSum ; proceed with sumval update
		
			nextADC_Ch0_valAboveLo:
			ldi temp1, LOW(NTC_HI - NTC_LO) ; load value for upper minus lower ADC limit (for NTC)
			ldi temp2, HIGH(NTC_HI - NTC_LO)
			cp adcSum0L, temp1 ; check if adcSum is beyond upper bound
			cpc adcSum0H, temp2
			brlo nextADC_Ch0_valWithinRange ; if ADC value is within setpoint range, continue with output calculation
				ldi setpointL, LOW(SP_MIN) ; load min. setpoint value
				ldi setpointH, HIGH(SP_MIN)
				rjmp nextADC_Ch0_clearSum ; proceed with sumval update

			nextADC_Ch0_valWithinRange:
			; calculate setpoint: setpoint = SP_MAX - ((((adcSum >> ADC_OS) - ADC_LO) * SP_SLOPE) >> 8),
			;	where ((adcSum >> ADC_OS) - ADC_LO) is already available in adcSum0H:L
			ldi setpointL, LOW(SP_SLOPE) ; load slope value (multiplicant), setpointH:L used as buffer
			ldi setpointH, HIGH(SP_SLOPE)
			mov temp2, adcSum0L ; load multiplicator (will be overwritten)
			mov bufL, adcSum0H ; bufL will then contain result divided by 256 (Lo byte)
			; call multiplication macro: result = SP_SLOPE * adcSum
			MUL_16_16_32 temp3, bufH, bufL, temp2, setpointH, setpointL, temp1
			; bufH:L contains multiplication result divided by 256
			; calculate setpointH:L = SP_MAX - bufH:L (negative slope):
			ldi setpointL, LOW(SP_MAX) 
			ldi setpointH, HIGH(SP_MAX)
			sub setpointL, bufL
			sbc setpointH, bufH 

			nextADC_Ch0_clearSum:
			clr adcSum0L ; reset ADC sum value
			clr adcSum0H

			rjmp nextADC_noAction

		nextADC_action_not0:
			cpi adcCh, 1 ; channel 1 (regulator input)?
			breq nextADC_action_1
				rjmp nextADC_action_not1 ; should never happen (for easy extension)
			nextADC_action_1:
				; Update regulator
				;
				; divide sumVal by 2^ADC_OS_BITS:
				; first add half of ADC_OS for rounding:
				ldi temp1, ADC_OS / 2
				add adcSum1L, temp1
				adc adcSum1H, zeroReg
				; now start division (cycles: 4 or 7 + 5 * (ADC_OS_BITS - 1) = 4, 7, 12, 17,..., max. 32)
				ldi temp1, ADC_OS_BITS ; init counter ([+1])
				tst temp1 ; ([+1])
				breq nextADC_Ch1_divide_1 ; ([+2]/[+1])
				nextADC_Ch1_divide_0:
					lsr adcSum1H ; divide by 2 (1 cyc [+1])
					ror adcSum1L ; (1 cyc [+1])
					dec temp1 ; decrease counter (1 cyc [+1])
					brne nextADC_Ch1_divide_0 ; (2 cyc [+1]]
				nextADC_Ch1_divide_1:

				; now calculate regulator error (setpoint - measured value)
				mov bufL, setpointL
				mov bufH, setpointH
				sub bufL, adcSum1L
				sbc bufH, adcSum1H

				ldi temp2, LOW(REG_I) ; load regulator parameter I
				ldi temp3, HIGH(REG_I)

				; call signed multiplication macro: result = error * regI
				MULS_16_16_32 buf1, buf0, bufH, bufL, temp3, temp2, temp1

				; add multiplication result to integral sum:
				add regI0, bufL
				adc regI1, bufH
				adc regI2, buf0
				adc regI3, buf1
				
				ldi temp1, PWM_MIN ; load boundaries for range checking (SREG remains unchanged)
				ldi temp2, PWM_MAX ; load boundaries for range checking (SREG remains unchanged)

				; check range:
				brpl nextADC_Ch1_regPos ; if integral sum is positive or zero continue with range checking
					; regI is negative -> set to min. value (and thus PWM output to PWM_MIN):
					rjmp nextADC_Ch1_regBelowMin

				nextADC_Ch1_regPos:  ; regI is not negative
					; compare to PWM_MIN (stored in temp1):
					cp regI0, zeroReg
					cpc regI1, zeroReg
					cpc regI2, temp1
					cpc regI3, zeroReg
					brcc nextADC_Ch1_regAboveMin ; continue with rage check if regI is not below minimum
				nextADC_Ch1_regBelowMin:
					; below minimum -> set to min. value (and thus PWM output to PWM_MIN):
					clr regI0
					clr regI1
					mov regI2, temp1 ; then integral sum contains value 2^16 * PWM_MIN
					clr regI3
					rjmp nextADC_Ch1_updatePWM ; update PWM output

				nextADC_Ch1_regAboveMin: ; regI is above minimum value
					; compare to PWM_MAX (stored in temp2):
					cp regI0, zeroReg
					cpc regI1, zeroReg
					cpc regI2, temp2
					cpc regI3, zeroReg
					brcs nextADC_Ch1_updatePWM ; update PWM output if below maximum value
					; regI is above maximum -> set to max. value (and thus PWM output to PWM_MAX):
					clr regI0
					clr regI1
					mov regI2, temp2 ; then integral sum contains value 2^16 * PWM_MAX
					clr regI3

				nextADC_Ch1_updatePWM:
				; update PWM value for compare match:
				out OCR1A, regI2 ; = regI / 2^16, highest byte regI3 is ignored
					
				nextADC_Ch1_clearSum:
				clr adcSum1L ; reset ADC sum value
				clr adcSum1H
				rjmp nextADC_noAction

		nextADC_action_not1:
			rjmp nextADC_action_not1 ; freeze (should never happen)


	nextADC_noAction:
		tst adcCh ; check channel (was already decremented)
		brne nextADC_ret ; if not zero, return from subprogram
			ldi adcCh, ADCNumChannels ; init channel counter
			dec osCnt ; decrease oversampling counter
			brne nextADC_ret ; if not zero, return from subprogram
				ldi osCnt, ADC_OS ; init counter (sumVal is cleared in nextADC_action)

	nextADC_ret:
	ret


; ============================================
; I N T E R R U P T   S E R V I C E S
; ============================================

isr_timer0_CMA: ; timer0 compare match A
	push temp1 ; save register
	in temp1, SREG ; save status register
	push temp1
	
	; set jobFlag:
	sbr jobFlag, (1<<0) ; changes status register

	pop temp1 ; get status register
	out SREG, temp1 ; restore status register
	pop temp1 ; restore register
	reti

; ============================================
; D A T A   S T O R A G E
; ============================================
; Strings, text, tables, etc.
; Format: 
; Label:
; .db "   example text   ",0



; End of source code
.db "(C)2015 by IB " ; human
.db "C(2)10 5ybI  B" ; word wise



